@@ -1,6 +1,6 @@
 # __init__.py
 
-from pyrmle import *
+from pyrmle.pyrmle import *
 
 # Version of the PyRMLE package
 __version__ = "1.0.0"
